@@ -18,12 +18,17 @@ MVMObject * MVM_6model_get_how_obj(MVMThreadContext *tc, MVMObject *obj) {
 static MVMObject * get_method_cache(MVMThreadContext *tc, MVMSTable *st) {
     if (!st->method_cache)
         MVM_serialization_finish_deserialize_method_cache(tc, st);
-   return st->method_cache;
+    return st->method_cache;
 }
 
 /* Locates a method by name, checking in the method cache only. */
 MVMObject * MVM_6model_find_method_cache_only(MVMThreadContext *tc, MVMObject *obj, MVMString *name) {
-    MVMObject *cache = get_method_cache(tc, STABLE(obj));
+    MVMObject *cache;
+
+    MVMROOT(tc, name, {
+        cache = get_method_cache(tc, STABLE(obj));
+    });
+
     if (cache && IS_CONCRETE(cache))
         return MVM_repr_at_key_o(tc, cache, name);
     return NULL;
@@ -52,7 +57,7 @@ static void die_over_missing_method(MVMThreadContext *tc, MVMObject *obj, MVMStr
         char *waste[] = { c_name, NULL };
         MVM_exception_throw_adhoc_free(tc, waste,
             "Cannot find method '%s' on object of type %s",
-            c_name, STABLE(obj)->debug_name);
+            c_name, MVM_6model_get_debug_name(tc, obj));
     }
 }
 static void late_bound_find_method_return(MVMThreadContext *tc, void *sr_data) {
@@ -68,13 +73,13 @@ static void late_bound_find_method_return(MVMThreadContext *tc, void *sr_data) {
     }
 }
 static void mark_find_method_sr_data(MVMThreadContext *tc, MVMFrame *frame, MVMGCWorklist *worklist) {
-    FindMethodSRData *fm = (FindMethodSRData *)frame->special_return_data;
+    FindMethodSRData *fm = (FindMethodSRData *)frame->extra->special_return_data;
     MVM_gc_worklist_add(tc, worklist, &fm->obj);
     MVM_gc_worklist_add(tc, worklist, &fm->name);
 }
 void MVM_6model_find_method(MVMThreadContext *tc, MVMObject *obj, MVMString *name, MVMRegister *res) {
-    MVMObject *cache, *HOW, *find_method, *code;
-    MVMCallsite *findmeth_callsite;
+    MVMObject *cache = NULL, *HOW = NULL, *find_method = NULL, *code = NULL;
+    MVMCallsite *findmeth_callsite = NULL;
 
     if (MVM_is_null(tc, obj)) {
         char *c_name  = MVM_string_utf8_encode_C_string(tc, name);
@@ -86,7 +91,10 @@ void MVM_6model_find_method(MVMThreadContext *tc, MVMObject *obj, MVMString *nam
 
     /* First try to find it in the cache. If we find it, we have a result.
      * If we don't find it, but the cache is authoritative, then error. */
-    cache = get_method_cache(tc, STABLE(obj));
+    MVMROOT2(tc, obj, name, {
+        cache = get_method_cache(tc, STABLE(obj));
+    });
+
     if (cache && IS_CONCRETE(cache)) {
         MVMObject *meth = MVM_repr_at_key_o(tc, cache, name);
         if (!MVM_is_null(tc, meth)) {
@@ -101,9 +109,12 @@ void MVM_6model_find_method(MVMThreadContext *tc, MVMObject *obj, MVMString *nam
 
     /* Otherwise, need to call the find_method method. We make the assumption
      * that the invocant's meta-object's type is composed. */
-    HOW = MVM_6model_get_how(tc, STABLE(obj));
-    find_method = MVM_6model_find_method_cache_only(tc, HOW,
-        tc->instance->str_consts.find_method);
+    MVMROOT3(tc, obj, name, HOW, {
+       HOW = MVM_6model_get_how(tc, STABLE(obj));
+       find_method = MVM_6model_find_method_cache_only(tc, HOW,
+            tc->instance->str_consts.find_method);
+    });
+
     if (MVM_is_null(tc, find_method)) {
         char *c_name  = MVM_string_utf8_encode_C_string(tc, name);
         char *waste[] = { c_name, NULL };
@@ -121,9 +132,8 @@ void MVM_6model_find_method(MVMThreadContext *tc, MVMObject *obj, MVMString *nam
         fm->obj  = obj;
         fm->name = name;
         fm->res  = res;
-        tc->cur_frame->special_return           = late_bound_find_method_return;
-        tc->cur_frame->special_return_data      = fm;
-        tc->cur_frame->mark_special_return_data = mark_find_method_sr_data;
+        MVM_frame_special_return(tc, tc->cur_frame, late_bound_find_method_return,
+            NULL, fm, mark_find_method_sr_data);
     }
     tc->cur_frame->args[0].o = HOW;
     tc->cur_frame->args[1].o = obj;
@@ -133,19 +143,26 @@ void MVM_6model_find_method(MVMThreadContext *tc, MVMObject *obj, MVMString *nam
 
 MVMint32 MVM_6model_find_method_spesh(MVMThreadContext *tc, MVMObject *obj, MVMString *name,
                                       MVMint32 ss_idx, MVMRegister *res) {
+    MVMObject *meth;
+
     /* Missed mono-morph; try cache-only lookup. */
-    MVMObject *meth = MVM_6model_find_method_cache_only(tc, obj, name);
+
+    MVMROOT2(tc, obj, name, {
+        meth = MVM_6model_find_method_cache_only(tc, obj, name);
+    });
+
     if (!MVM_is_null(tc, meth)) {
         /* Got it; cache. Must be careful due to threads
          * reading, races, etc. */
         MVMStaticFrame *sf = tc->cur_frame->static_info;
         uv_mutex_lock(&tc->instance->mutex_spesh_install);
         if (!tc->cur_frame->effective_spesh_slots[ss_idx + 1]) {
-            MVM_ASSIGN_REF(tc, &(sf->common.header),
+            MVMStaticFrameSpesh *spesh = sf->body.spesh;
+            MVM_ASSIGN_REF(tc, &(spesh->common.header),
                            tc->cur_frame->effective_spesh_slots[ss_idx + 1],
                            (MVMCollectable *)meth);
             MVM_barrier();
-            MVM_ASSIGN_REF(tc, &(sf->common.header),
+            MVM_ASSIGN_REF(tc, &(spesh->common.header),
                            tc->cur_frame->effective_spesh_slots[ss_idx],
                            (MVMCollectable *)STABLE(obj));
         }
@@ -180,7 +197,11 @@ MVMint64 MVM_6model_can_method_cache_only(MVMThreadContext *tc, MVMObject *obj,
     }
 
     /* Consider the method cache. */
-    cache = get_method_cache(tc, STABLE(obj));
+
+    MVMROOT2(tc, obj, name, {
+        cache = get_method_cache(tc, STABLE(obj));
+    });
+
     if (cache && IS_CONCRETE(cache)) {
         MVMObject *meth = MVM_repr_at_key_o(tc, cache, name);
         if (!MVM_is_null(tc, meth)) {
@@ -194,10 +215,14 @@ MVMint64 MVM_6model_can_method_cache_only(MVMThreadContext *tc, MVMObject *obj,
 }
 
 void MVM_6model_can_method(MVMThreadContext *tc, MVMObject *obj, MVMString *name, MVMRegister *res) {
-    MVMObject *HOW, *find_method, *code;
-    MVMCallsite *findmeth_callsite;
+    MVMObject *HOW = NULL, *find_method = NULL, *code = NULL;
+    MVMCallsite *findmeth_callsite = NULL;
 
-    MVMint64 can_cached = MVM_6model_can_method_cache_only(tc, obj, name);
+    MVMint64 can_cached;
+
+    MVMROOT2(tc, obj, name, {
+        can_cached = MVM_6model_can_method_cache_only(tc, obj, name);
+    });
 
     if (can_cached == 0 || can_cached == 1) {
         res->i64 = can_cached;
@@ -206,9 +231,12 @@ void MVM_6model_can_method(MVMThreadContext *tc, MVMObject *obj, MVMString *name
 
     /* If no method in cache and the cache is not authoritative, need to make
      * a late-bound call to find_method. */
-    HOW = MVM_6model_get_how(tc, STABLE(obj));
-    find_method = MVM_6model_find_method_cache_only(tc, HOW,
-        tc->instance->str_consts.find_method);
+    MVMROOT3(tc, obj, name, HOW, {
+        HOW = MVM_6model_get_how(tc, STABLE(obj));
+        find_method = MVM_6model_find_method_cache_only(tc, HOW,
+            tc->instance->str_consts.find_method);
+    });
+
     if (MVM_is_null(tc, find_method)) {
         /* This'll count as a "no"... */
         res->i64 = 0;
@@ -220,8 +248,7 @@ void MVM_6model_can_method(MVMThreadContext *tc, MVMObject *obj, MVMString *name
     code = MVM_frame_find_invokee(tc, find_method, NULL);
     findmeth_callsite = MVM_callsite_get_common(tc, MVM_CALLSITE_ID_FIND_METHOD);
     MVM_args_setup_thunk(tc, res, MVM_RETURN_OBJ, findmeth_callsite);
-    tc->cur_frame->special_return      = late_bound_can_return;
-    tc->cur_frame->special_return_data = res;
+    MVM_frame_special_return(tc, tc->cur_frame, late_bound_can_return, NULL, res, NULL);
     tc->cur_frame->args[0].o = HOW;
     tc->cur_frame->args[1].o = obj;
     tc->cur_frame->args[2].s = name;
@@ -231,9 +258,14 @@ void MVM_6model_can_method(MVMThreadContext *tc, MVMObject *obj, MVMString *name
 /* Checks if an object has a given type, delegating to the type_check or
  * accepts_type methods as needed. */
 static void do_accepts_type_check(MVMThreadContext *tc, MVMObject *obj, MVMObject *type, MVMRegister *res) {
-    MVMObject *HOW = MVM_6model_get_how(tc, STABLE(type));
-    MVMObject *meth = MVM_6model_find_method_cache_only(tc, HOW,
-        tc->instance->str_consts.accepts_type);
+    MVMObject *HOW = NULL, *meth = NULL;
+
+    MVMROOT3(tc, obj, type, HOW, {
+        HOW = MVM_6model_get_how(tc, STABLE(type));
+        meth = MVM_6model_find_method_cache_only(tc, HOW,
+            tc->instance->str_consts.accepts_type);
+    });
+
     if (!MVM_is_null(tc, meth)) {
         /* Set up the call, using the result register as the target. */
         MVMObject *code = MVM_frame_find_invokee(tc, meth, NULL);
@@ -268,7 +300,7 @@ static void accepts_type_sr(MVMThreadContext *tc, void *sr_data) {
 }
 
 static void mark_sr_data(MVMThreadContext *tc, MVMFrame *frame, MVMGCWorklist *worklist) {
-    AcceptsTypeSRData *atd = (AcceptsTypeSRData *)frame->special_return_data;
+    AcceptsTypeSRData *atd = (AcceptsTypeSRData *)frame->extra->special_return_data;
     MVM_gc_worklist_add(tc, worklist, &atd->obj);
     MVM_gc_worklist_add(tc, worklist, &atd->type);
 }
@@ -310,9 +342,13 @@ void MVM_6model_istype(MVMThreadContext *tc, MVMObject *obj, MVMObject *type, MV
     /* If we get here, need to call .^type_check on the value we're
      * checking, unless it's an accepts check. */
     if (!cache || (mode & MVM_TYPE_CHECK_CACHE_THEN_METHOD)) {
-        MVMObject *HOW = MVM_6model_get_how(tc, st);
-        MVMObject *meth = MVM_6model_find_method_cache_only(tc, HOW,
-            tc->instance->str_consts.type_check);
+        MVMObject *HOW = NULL, *meth = NULL;
+
+        MVMROOT3(tc, obj, type, HOW, {
+            HOW = MVM_6model_get_how(tc, st);
+            meth = MVM_6model_find_method_cache_only(tc, HOW,
+                tc->instance->str_consts.type_check);
+        });
         if (!MVM_is_null(tc, meth)) {
             /* Set up the call, using the result register as the target. */
             MVMObject *code = MVM_frame_find_invokee(tc, meth, NULL);
@@ -327,9 +363,8 @@ void MVM_6model_istype(MVMThreadContext *tc, MVMObject *obj, MVMObject *type, MV
                 atd->obj = obj;
                 atd->type = type;
                 atd->res = res;
-                tc->cur_frame->special_return           = accepts_type_sr;
-                tc->cur_frame->special_return_data      = atd;
-                tc->cur_frame->mark_special_return_data = mark_sr_data;
+                MVM_frame_special_return(tc, tc->cur_frame, accepts_type_sr, NULL,
+                    atd, mark_sr_data);
             }
             STABLE(code)->invoke(tc, code, typecheck_callsite, tc->cur_frame->args);
             return;
@@ -389,7 +424,7 @@ MVMint64 MVM_6model_try_cache_type_check(MVMThreadContext *tc, MVMObject *obj, M
 
 /* Default invoke function on STables; for non-invokable objects */
 void MVM_6model_invoke_default(MVMThreadContext *tc, MVMObject *invokee, MVMCallsite *callsite, MVMRegister *args) {
-    MVM_exception_throw_adhoc(tc, "Cannot invoke this object (REPR: %s; %s)", REPR(invokee)->name, STABLE(invokee)->debug_name);
+    MVM_exception_throw_adhoc(tc, "Cannot invoke this object (REPR: %s; %s)", REPR(invokee)->name, MVM_6model_get_debug_name(tc, invokee));
 }
 
 /* Clean up STable memory. */
