@@ -28,10 +28,10 @@ static void new_type(MVMThreadContext *tc, MVMCallsite *callsite, MVMRegister *a
     MVMInstance       *instance = tc->instance;
 
     /* Get arguments. */
-    MVMArgProcContext arg_ctx; arg_ctx.named_used = NULL;
+    MVMArgProcContext arg_ctx;
     MVM_args_proc_init(tc, &arg_ctx, callsite, args);
     MVM_args_checkarity(tc, &arg_ctx, 1, 1);
-    self = MVM_args_get_pos_obj(tc, &arg_ctx, 0, MVM_ARG_REQUIRED).arg.o;
+    self = MVM_args_get_required_pos_obj(tc, &arg_ctx, 0);
     repr_arg = MVM_args_get_named_str(tc, &arg_ctx, instance->str_consts.repr, MVM_ARG_OPTIONAL);
     name_arg = MVM_args_get_named_str(tc, &arg_ctx, instance->str_consts.name, MVM_ARG_OPTIONAL);
     MVM_args_proc_cleanup(tc, &arg_ctx);
@@ -79,12 +79,12 @@ static void add_method(MVMThreadContext *tc, MVMCallsite *callsite, MVMRegister
     MVMString *name;
 
     /* Get arguments. */
-    MVMArgProcContext arg_ctx; arg_ctx.named_used = NULL;
+    MVMArgProcContext arg_ctx;
     MVM_args_proc_init(tc, &arg_ctx, callsite, args);
     MVM_args_checkarity(tc, &arg_ctx, 4, 4);
-    self     = MVM_args_get_pos_obj(tc, &arg_ctx, 0, MVM_ARG_REQUIRED).arg.o;
-    name     = MVM_args_get_pos_str(tc, &arg_ctx, 2, MVM_ARG_REQUIRED).arg.s;
-    method   = MVM_args_get_pos_obj(tc, &arg_ctx, 3, MVM_ARG_REQUIRED).arg.o;
+    self     = MVM_args_get_required_pos_obj(tc, &arg_ctx, 0);
+    name     = MVM_args_get_required_pos_str(tc, &arg_ctx, 2);
+    method   = MVM_args_get_required_pos_obj(tc, &arg_ctx, 3);
     MVM_args_proc_cleanup(tc, &arg_ctx);
     if (!self || !IS_CONCRETE(self) || REPR(self)->ID != MVM_REPR_ID_KnowHOWREPR)
         MVM_exception_throw_adhoc(tc, "KnowHOW methods must be called on object instance with REPR KnowHOWREPR");
@@ -102,11 +102,11 @@ static void add_attribute(MVMThreadContext *tc, MVMCallsite *callsite, MVMRegist
     MVMObject *self, *attr, *attributes;
 
     /* Get arguments. */
-    MVMArgProcContext arg_ctx; arg_ctx.named_used = NULL;
+    MVMArgProcContext arg_ctx;
     MVM_args_proc_init(tc, &arg_ctx, callsite, args);
     MVM_args_checkarity(tc, &arg_ctx, 3, 3);
-    self     = MVM_args_get_pos_obj(tc, &arg_ctx, 0, MVM_ARG_REQUIRED).arg.o;
-    attr     = MVM_args_get_pos_obj(tc, &arg_ctx, 2, MVM_ARG_REQUIRED).arg.o;
+    self     = MVM_args_get_required_pos_obj(tc, &arg_ctx, 0);
+    attr     = MVM_args_get_required_pos_obj(tc, &arg_ctx, 2);
     MVM_args_proc_cleanup(tc, &arg_ctx);
 
     /* Ensure we have the required representations. */
@@ -131,11 +131,11 @@ static void compose(MVMThreadContext *tc, MVMCallsite *callsite, MVMRegister *ar
     MVMInstance *instance = tc->instance;
 
     /* Get arguments. */
-    MVMArgProcContext arg_ctx; arg_ctx.named_used = NULL;
+    MVMArgProcContext arg_ctx;
     MVM_args_proc_init(tc, &arg_ctx, callsite, args);
     MVM_args_checkarity(tc, &arg_ctx, 2, 2);
-    self     = MVM_args_get_pos_obj(tc, &arg_ctx, 0, MVM_ARG_REQUIRED).arg.o;
-    type_obj = MVM_args_get_pos_obj(tc, &arg_ctx, 1, MVM_ARG_REQUIRED).arg.o;
+    self     = MVM_args_get_required_pos_obj(tc, &arg_ctx, 0);
+    type_obj = MVM_args_get_required_pos_obj(tc, &arg_ctx, 1);
     MVM_args_proc_cleanup(tc, &arg_ctx);
     if (!self || !IS_CONCRETE(self) || REPR(self)->ID != MVM_REPR_ID_KnowHOWREPR)
         MVM_exception_throw_adhoc(tc, "KnowHOW methods must be called on object instance with REPR KnowHOWREPR");
@@ -181,21 +181,19 @@ static void compose(MVMThreadContext *tc, MVMCallsite *callsite, MVMRegister *ar
         MVMObject *attr_info = REPR(BOOTHash)->allocate(tc, STABLE(BOOTHash));
         MVMKnowHOWAttributeREPR *attribute = (MVMKnowHOWAttributeREPR *)
             MVM_repr_at_pos_o(tc, attributes, i);
-        MVMROOT(tc, attr_info, {
-            MVMROOT(tc, attribute, {
-                if (REPR((MVMObject *)attribute)->ID != MVM_REPR_ID_KnowHOWAttributeREPR)
-                    MVM_exception_throw_adhoc(tc, "KnowHOW attributes must use KnowHOWAttributeREPR");
+        MVMROOT2(tc, attr_info, attribute, {
+            if (REPR((MVMObject *)attribute)->ID != MVM_REPR_ID_KnowHOWAttributeREPR)
+                MVM_exception_throw_adhoc(tc, "KnowHOW attributes must use KnowHOWAttributeREPR");
 
-                MVM_repr_init(tc, attr_info);
-                MVM_repr_bind_key_o(tc, attr_info, instance->str_consts.name, (MVMObject *)attribute->body.name);
-                MVM_repr_bind_key_o(tc, attr_info, instance->str_consts.type, attribute->body.type);
-                if (attribute->body.box_target) {
-                    /* Merely having the key serves as a "yes". */
-                    MVM_repr_bind_key_o(tc, attr_info, instance->str_consts.box_target, attr_info);
-                }
+            MVM_repr_init(tc, attr_info);
+            MVM_repr_bind_key_o(tc, attr_info, instance->str_consts.name, (MVMObject *)attribute->body.name);
+            MVM_repr_bind_key_o(tc, attr_info, instance->str_consts.type, attribute->body.type);
+            if (attribute->body.box_target) {
+                /* Merely having the key serves as a "yes". */
+                MVM_repr_bind_key_o(tc, attr_info, instance->str_consts.box_target, attr_info);
+            }
 
-                MVM_repr_push_o(tc, attr_info_list, attr_info);
-            });
+            MVM_repr_push_o(tc, attr_info_list, attr_info);
         });
     }
 
@@ -224,11 +222,11 @@ static void compose(MVMThreadContext *tc, MVMCallsite *callsite, MVMRegister *ar
 #define introspect_member(member, set_result, result) \
 static void member(MVMThreadContext *tc, MVMCallsite *callsite, MVMRegister *args) { \
     MVMObject *self, *type_obj, *member; \
-    MVMArgProcContext arg_ctx; arg_ctx.named_used = NULL; \
+    MVMArgProcContext arg_ctx; \
     MVM_args_proc_init(tc, &arg_ctx, callsite, args); \
     MVM_args_checkarity(tc, &arg_ctx, 2, 2); \
-    self     = MVM_args_get_pos_obj(tc, &arg_ctx, 0, MVM_ARG_REQUIRED).arg.o; \
-    type_obj = MVM_args_get_pos_obj(tc, &arg_ctx, 1, MVM_ARG_REQUIRED).arg.o; \
+    self     = MVM_args_get_required_pos_obj(tc, &arg_ctx, 0); \
+    type_obj = MVM_args_get_required_pos_obj(tc, &arg_ctx, 1); \
     MVM_args_proc_cleanup(tc, &arg_ctx); \
     if (!self || !IS_CONCRETE(self) || REPR(self)->ID != MVM_REPR_ID_KnowHOWREPR) \
         MVM_exception_throw_adhoc(tc, "KnowHOW methods must be called on object instance with REPR KnowHOWREPR"); \
@@ -340,10 +338,10 @@ static void attr_new(MVMThreadContext *tc, MVMCallsite *callsite, MVMRegister *a
     MVMInstance       *instance = tc->instance;
 
     /* Process arguments. */
-    MVMArgProcContext arg_ctx; arg_ctx.named_used = NULL;
+    MVMArgProcContext arg_ctx;
     MVM_args_proc_init(tc, &arg_ctx, callsite, args);
     MVM_args_checkarity(tc, &arg_ctx, 1, 1);
-    self     = MVM_args_get_pos_obj(tc, &arg_ctx, 0, MVM_ARG_REQUIRED).arg.o;
+    self     = MVM_args_get_required_pos_obj(tc, &arg_ctx, 0);
     name_arg = MVM_args_get_named_str(tc, &arg_ctx, instance->str_consts.name, MVM_ARG_REQUIRED);
     type_arg = MVM_args_get_named_obj(tc, &arg_ctx, instance->str_consts.type, MVM_ARG_OPTIONAL);
     bt_arg   = MVM_args_get_named_int(tc, &arg_ctx, instance->str_consts.box_target, MVM_ARG_OPTIONAL);
@@ -370,10 +368,10 @@ static void attr_new(MVMThreadContext *tc, MVMCallsite *callsite, MVMRegister *a
 /* Composes the attribute; actually, nothing to do really. */
 static void attr_compose(MVMThreadContext *tc, MVMCallsite *callsite, MVMRegister *args) {
     MVMObject *self;
-    MVMArgProcContext arg_ctx; arg_ctx.named_used = NULL;
+    MVMArgProcContext arg_ctx;
     MVM_args_proc_init(tc, &arg_ctx, callsite, args);
     MVM_args_checkarity(tc, &arg_ctx, 1, 1);
-    self = MVM_args_get_pos_obj(tc, &arg_ctx, 0, MVM_ARG_REQUIRED).arg.o;
+    self = MVM_args_get_required_pos_obj(tc, &arg_ctx, 0);
     MVM_args_proc_cleanup(tc, &arg_ctx);
     MVM_args_set_result_obj(tc, self, MVM_RETURN_CURRENT_FRAME);
 }
@@ -382,10 +380,10 @@ static void attr_compose(MVMThreadContext *tc, MVMCallsite *callsite, MVMRegiste
 static void attr_name(MVMThreadContext *tc, MVMCallsite *callsite, MVMRegister *args) {
     MVMObject *self;
     MVMString *name;
-    MVMArgProcContext arg_ctx; arg_ctx.named_used = NULL;
+    MVMArgProcContext arg_ctx;
     MVM_args_proc_init(tc, &arg_ctx, callsite, args);
     MVM_args_checkarity(tc, &arg_ctx, 1, 1);
-    self = MVM_args_get_pos_obj(tc, &arg_ctx, 0, MVM_ARG_REQUIRED).arg.o;
+    self = MVM_args_get_required_pos_obj(tc, &arg_ctx, 0);
     MVM_args_proc_cleanup(tc, &arg_ctx);
     name = ((MVMKnowHOWAttributeREPR *)self)->body.name;
     MVM_args_set_result_str(tc, name, MVM_RETURN_CURRENT_FRAME);
@@ -394,10 +392,10 @@ static void attr_name(MVMThreadContext *tc, MVMCallsite *callsite, MVMRegister *
 /* Introspects the attribute's type. */
 static void attr_type(MVMThreadContext *tc, MVMCallsite *callsite, MVMRegister *args) {
     MVMObject *self, *type;
-    MVMArgProcContext arg_ctx; arg_ctx.named_used = NULL;
+    MVMArgProcContext arg_ctx;
     MVM_args_proc_init(tc, &arg_ctx, callsite, args);
     MVM_args_checkarity(tc, &arg_ctx, 1, 1);
-    self = MVM_args_get_pos_obj(tc, &arg_ctx, 0, MVM_ARG_REQUIRED).arg.o;
+    self = MVM_args_get_required_pos_obj(tc, &arg_ctx, 0);
     MVM_args_proc_cleanup(tc, &arg_ctx);
     type = ((MVMKnowHOWAttributeREPR *)self)->body.type;
     MVM_args_set_result_obj(tc, type, MVM_RETURN_CURRENT_FRAME);
@@ -407,10 +405,10 @@ static void attr_type(MVMThreadContext *tc, MVMCallsite *callsite, MVMRegister *
 static void attr_box_target(MVMThreadContext *tc, MVMCallsite *callsite, MVMRegister *args) {
     MVMObject *self;
     MVMint64   box_target;
-    MVMArgProcContext arg_ctx; arg_ctx.named_used = NULL;
+    MVMArgProcContext arg_ctx;
     MVM_args_proc_init(tc, &arg_ctx, callsite, args);
     MVM_args_checkarity(tc, &arg_ctx, 1, 1);
-    self = MVM_args_get_pos_obj(tc, &arg_ctx, 0, MVM_ARG_REQUIRED).arg.o;
+    self = MVM_args_get_required_pos_obj(tc, &arg_ctx, 0);
     MVM_args_proc_cleanup(tc, &arg_ctx);
     box_target = ((MVMKnowHOWAttributeREPR *)self)->body.box_target;
     MVM_args_set_result_int(tc, box_target, MVM_RETURN_CURRENT_FRAME);
@@ -582,6 +580,7 @@ static void string_consts(MVMThreadContext *tc) {
     string_creator(heap, "heap");
     string_creator(translate_newlines, "translate_newlines");
     string_creator(platform_newline, MVM_TRANSLATE_NEWLINE_OUTPUT ? "\r\n" : "\n");
+    string_creator(path, "path");
 }
 
 /* Drives the overall bootstrap process. */
@@ -598,7 +597,7 @@ void MVM_6model_bootstrap(MVMThreadContext *tc) {
     MVM_repr_initialize_registry(tc);
 
     /* Create stub VMNull, BOOTInt, BOOTNum, BOOTStr, BOOTArray, BOOTHash,
-     * BOOTCCode, BOOTCode, BOOTThread, BOOTIter, BOOTContext, SCRef, Lexotic,
+     * BOOTCCode, BOOTCode, BOOTThread, BOOTIter, BOOTContext, SCRef,
      * CallCapture, BOOTIO, BOOTException, BOOTQueue, BOOTAsync,
      * and BOOTReentrantMutex types. */
 #define create_stub_boot_type(tc, reprid, slot, makeboolspec, boolspec) do { \
@@ -624,7 +623,6 @@ void MVM_6model_bootstrap(MVMThreadContext *tc) {
     create_stub_boot_type(tc, MVM_REPR_ID_MVMIter, boot_types.BOOTIter, 1, MVM_BOOL_MODE_ITER);
     create_stub_boot_type(tc, MVM_REPR_ID_MVMContext, boot_types.BOOTContext, 0, MVM_BOOL_MODE_NOT_TYPE_OBJECT);
     create_stub_boot_type(tc, MVM_REPR_ID_SCRef, SCRef, 0, MVM_BOOL_MODE_NOT_TYPE_OBJECT);
-    create_stub_boot_type(tc, MVM_REPR_ID_Lexotic, Lexotic, 0, MVM_BOOL_MODE_NOT_TYPE_OBJECT);
     create_stub_boot_type(tc, MVM_REPR_ID_MVMCallCapture, CallCapture, 0, MVM_BOOL_MODE_NOT_TYPE_OBJECT);
     create_stub_boot_type(tc, MVM_REPR_ID_MVMOSHandle, boot_types.BOOTIO, 0, MVM_BOOL_MODE_NOT_TYPE_OBJECT);
     create_stub_boot_type(tc, MVM_REPR_ID_MVMException, boot_types.BOOTException, 0, MVM_BOOL_MODE_NOT_TYPE_OBJECT);
@@ -636,6 +634,8 @@ void MVM_6model_bootstrap(MVMThreadContext *tc) {
     create_stub_boot_type(tc, MVM_REPR_ID_ConcBlockingQueue, boot_types.BOOTQueue, 0, MVM_BOOL_MODE_NOT_TYPE_OBJECT);
     create_stub_boot_type(tc, MVM_REPR_ID_MVMAsyncTask, boot_types.BOOTAsync, 0, MVM_BOOL_MODE_NOT_TYPE_OBJECT);
     create_stub_boot_type(tc, MVM_REPR_ID_ReentrantMutex, boot_types.BOOTReentrantMutex, 0, MVM_BOOL_MODE_NOT_TYPE_OBJECT);
+    create_stub_boot_type(tc, MVM_REPR_ID_MVMSpeshLog, SpeshLog, 0, MVM_BOOL_MODE_NOT_TYPE_OBJECT);
+    create_stub_boot_type(tc, MVM_REPR_ID_MVMStaticFrameSpesh, StaticFrameSpesh, 0, MVM_BOOL_MODE_NOT_TYPE_OBJECT);
 
     /* Bootstrap the KnowHOW type, giving it a meta-object. */
     bootstrap_KnowHOW(tc);
@@ -658,7 +658,6 @@ void MVM_6model_bootstrap(MVMThreadContext *tc) {
     meta_objectifier(tc, boot_types.BOOTIter, "BOOTIter");
     meta_objectifier(tc, boot_types.BOOTContext, "BOOTContext");
     meta_objectifier(tc, SCRef, "SCRef");
-    meta_objectifier(tc, Lexotic, "Lexotic");
     meta_objectifier(tc, CallCapture, "CallCapture");
     meta_objectifier(tc, boot_types.BOOTIO, "BOOTIO");
     meta_objectifier(tc, boot_types.BOOTException, "BOOTException");
