@@ -179,7 +179,7 @@ static void * unmarshal_callback(MVMThreadContext *tc, MVMObject *callback, MVMO
 
         MVM_callsite_try_intern(tc, &cs);
 
-        callback_data->tc        = tc;
+        callback_data->instance  = tc->instance;
         callback_data->cs        = cs;
         callback_data->target    = callback;
         callback_data->cb        = dcbNewCallback(signature, (DCCallbackHandler *)callback_handler, callback_data);
@@ -213,8 +213,10 @@ static char callback_handler(DCCallback *cb, DCArgs *cb_args, DCValue *cb_result
     MVMRegister *args;
     unsigned int interval_id;
 
+    /* Locate the MoarVM thread this callback is being run on. */
+    MVMThreadContext *tc = MVM_nativecall_find_thread_context(data->instance);
+
     /* Unblock GC if needed, so this thread can do work. */
-    MVMThreadContext *tc = data->tc;
     MVMint32 was_blocked = MVM_gc_is_thread_blocked(tc);
     if (was_blocked)
         MVM_gc_mark_thread_unblocked(tc);
@@ -324,8 +326,7 @@ static char callback_handler(DCCallback *cb, DCArgs *cb_args, DCValue *cb_result
         MVMCompUnit **backup_interp_cu          = tc->interp_cu;
         MVMFrame *backup_cur_frame              = MVM_frame_force_to_heap(tc, tc->cur_frame);
         MVMFrame *backup_thread_entry_frame     = tc->thread_entry_frame;
-        MVMROOT(tc, backup_cur_frame, {
-        MVMROOT(tc, backup_thread_entry_frame, {
+        MVMROOT2(tc, backup_cur_frame, backup_thread_entry_frame, {
             MVMuint32 backup_mark                   = MVM_gc_root_temp_mark(tc);
             jmp_buf backup_interp_jump;
             memcpy(backup_interp_jump, tc->interp_jump, sizeof(jmp_buf));
@@ -344,80 +345,79 @@ static char callback_handler(DCCallback *cb, DCArgs *cb_args, DCValue *cb_result
             memcpy(tc->interp_jump, backup_interp_jump, sizeof(jmp_buf));
             MVM_gc_root_temp_mark_reset(tc, backup_mark);
         });
-        });
     }
 
     /* Handle return value. */
     if (res.o) {
         MVMContainerSpec const *contspec = STABLE(res.o)->container_spec;
         if (contspec && contspec->fetch_never_invokes)
-            contspec->fetch(data->tc, res.o, &res);
+            contspec->fetch(tc, res.o, &res);
     }
     switch (data->typeinfos[0] & MVM_NATIVECALL_ARG_TYPE_MASK) {
         case MVM_NATIVECALL_ARG_VOID:
             break;
         case MVM_NATIVECALL_ARG_CHAR:
-            cb_result->c = (signed char)MVM_nativecall_unmarshal_char(data->tc, res.o);
+            cb_result->c = (signed char)MVM_nativecall_unmarshal_char(tc, res.o);
             break;
         case MVM_NATIVECALL_ARG_SHORT:
-            cb_result->s = MVM_nativecall_unmarshal_short(data->tc, res.o);
+            cb_result->s = MVM_nativecall_unmarshal_short(tc, res.o);
             break;
         case MVM_NATIVECALL_ARG_INT:
-            cb_result->i = MVM_nativecall_unmarshal_int(data->tc, res.o);
+            cb_result->i = MVM_nativecall_unmarshal_int(tc, res.o);
             break;
         case MVM_NATIVECALL_ARG_LONG:
-            cb_result->j = MVM_nativecall_unmarshal_long(data->tc, res.o);
+            cb_result->j = MVM_nativecall_unmarshal_long(tc, res.o);
             break;
         case MVM_NATIVECALL_ARG_LONGLONG:
-            cb_result->l = MVM_nativecall_unmarshal_longlong(data->tc, res.o);
+            cb_result->l = MVM_nativecall_unmarshal_longlong(tc, res.o);
             break;
         case MVM_NATIVECALL_ARG_FLOAT:
-            cb_result->f = MVM_nativecall_unmarshal_float(data->tc, res.o);
+            cb_result->f = MVM_nativecall_unmarshal_float(tc, res.o);
             break;
         case MVM_NATIVECALL_ARG_DOUBLE:
-            cb_result->d = MVM_nativecall_unmarshal_double(data->tc, res.o);
+            cb_result->d = MVM_nativecall_unmarshal_double(tc, res.o);
             break;
         case MVM_NATIVECALL_ARG_ASCIISTR:
         case MVM_NATIVECALL_ARG_UTF8STR:
         case MVM_NATIVECALL_ARG_UTF16STR:
-            cb_result->Z = MVM_nativecall_unmarshal_string(data->tc, res.o, data->typeinfos[0], NULL);
+            cb_result->Z = MVM_nativecall_unmarshal_string(tc, res.o, data->typeinfos[0], NULL);
             break;
         case MVM_NATIVECALL_ARG_CSTRUCT:
-            cb_result->p = MVM_nativecall_unmarshal_cstruct(data->tc, res.o);
+            cb_result->p = MVM_nativecall_unmarshal_cstruct(tc, res.o);
             break;
         case MVM_NATIVECALL_ARG_CPOINTER:
-            cb_result->p = MVM_nativecall_unmarshal_cpointer(data->tc, res.o);
+            cb_result->p = MVM_nativecall_unmarshal_cpointer(tc, res.o);
             break;
         case MVM_NATIVECALL_ARG_CARRAY:
-            cb_result->p = MVM_nativecall_unmarshal_carray(data->tc, res.o);
+            cb_result->p = MVM_nativecall_unmarshal_carray(tc, res.o);
             break;
         case MVM_NATIVECALL_ARG_CUNION:
-            cb_result->p = MVM_nativecall_unmarshal_cunion(data->tc, res.o);
+            cb_result->p = MVM_nativecall_unmarshal_cunion(tc, res.o);
             break;
         case MVM_NATIVECALL_ARG_VMARRAY:
-            cb_result->p = MVM_nativecall_unmarshal_vmarray(data->tc, res.o);
+            cb_result->p = MVM_nativecall_unmarshal_vmarray(tc, res.o);
             break;
         case MVM_NATIVECALL_ARG_CALLBACK:
-            cb_result->p = unmarshal_callback(data->tc, res.o, data->types[0]);
+            cb_result->p = unmarshal_callback(tc, res.o, data->types[0]);
             break;
         case MVM_NATIVECALL_ARG_UCHAR:
-            cb_result->c = MVM_nativecall_unmarshal_uchar(data->tc, res.o);
+            cb_result->c = MVM_nativecall_unmarshal_uchar(tc, res.o);
             break;
         case MVM_NATIVECALL_ARG_USHORT:
-            cb_result->s = MVM_nativecall_unmarshal_ushort(data->tc, res.o);
+            cb_result->s = MVM_nativecall_unmarshal_ushort(tc, res.o);
             break;
         case MVM_NATIVECALL_ARG_UINT:
-            cb_result->i = MVM_nativecall_unmarshal_uint(data->tc, res.o);
+            cb_result->i = MVM_nativecall_unmarshal_uint(tc, res.o);
             break;
         case MVM_NATIVECALL_ARG_ULONG:
-            cb_result->j = MVM_nativecall_unmarshal_ulong(data->tc, res.o);
+            cb_result->j = MVM_nativecall_unmarshal_ulong(tc, res.o);
             break;
         case MVM_NATIVECALL_ARG_ULONGLONG:
-            cb_result->l = MVM_nativecall_unmarshal_ulonglong(data->tc, res.o);
+            cb_result->l = MVM_nativecall_unmarshal_ulonglong(tc, res.o);
             break;
         default:
             MVM_telemetry_interval_stop(tc, interval_id, "nativecall callback handler failed");
-            MVM_exception_throw_adhoc(data->tc,
+            MVM_exception_throw_adhoc(tc,
                 "Internal error: unhandled dyncall callback return type");
     }
 
@@ -483,9 +483,10 @@ MVMObject * MVM_nativecall_invoke(MVMThreadContext *tc, MVMObject *res_type,
     void     *ptr         = NULL;
 
     unsigned int interval_id;
+    DCCallVM *vm;
 
     /* Create and set up call VM. */
-    DCCallVM *vm = dcNewCallVM(8192);
+    vm = dcNewCallVM(8192);
     dcMode(vm, body->convention);
     dcReset(vm);
 
@@ -597,8 +598,7 @@ MVMObject * MVM_nativecall_invoke(MVMThreadContext *tc, MVMObject *res_type,
         }
     }
 
-    MVMROOT(tc, args, {
-    MVMROOT(tc, res_type, {
+    MVMROOT2(tc, args, res_type, {
         MVM_gc_mark_thread_blocked(tc);
         if (result) {
             /* We are calling a C++ constructor so we hand back the invocant (THIS) we recorded earlier. */
@@ -738,7 +738,6 @@ MVMObject * MVM_nativecall_invoke(MVMThreadContext *tc, MVMObject *res_type,
             }
         }
     });
-    });
 
     num_rws = 0;
     for (i = 0; i < num_args; i++) {
