@@ -75,10 +75,19 @@ void MVM_gc_root_add_instance_roots_to_worklist(MVMThreadContext *tc, MVMGCWorkl
     add_collectable(tc, worklist, snapshot, tc->instance->compiler_registry, "Compiler registry");
     add_collectable(tc, worklist, snapshot, tc->instance->hll_syms, "HLL symbols");
     add_collectable(tc, worklist, snapshot, tc->instance->clargs, "Command line args");
-    add_collectable(tc, worklist, snapshot, tc->instance->event_loop_todo_queue, "Event loop todo queue");
-    add_collectable(tc, worklist, snapshot, tc->instance->event_loop_cancel_queue, "Event loop cancel queue");
+    add_collectable(tc, worklist, snapshot, tc->instance->event_loop_todo_queue,
+        "Event loop todo queue");
+    add_collectable(tc, worklist, snapshot, tc->instance->event_loop_permit_queue,
+        "Event loop permit queue");
+    add_collectable(tc, worklist, snapshot, tc->instance->event_loop_cancel_queue,
+        "Event loop cancel queue");
     add_collectable(tc, worklist, snapshot, tc->instance->event_loop_active, "Event loop active");
 
+    add_collectable(tc, worklist, snapshot, tc->instance->spesh_queue,
+        "Specialization log queue");
+    if (worklist)
+        MVM_spesh_plan_gc_mark(tc, tc->instance->spesh_plan, worklist);
+
     int_to_str_cache = tc->instance->int_to_str_cache;
     for (i = 0; i < MVM_INT_TO_STR_CACHE_SIZE; i++)
         add_collectable(tc, worklist, snapshot, int_to_str_cache[i],
@@ -107,6 +116,9 @@ void MVM_gc_root_add_instance_roots_to_worklist(MVMThreadContext *tc, MVMGCWorkl
 
     add_collectable(tc, worklist, snapshot, tc->instance->cached_backend_config,
         "Cached backend configuration hash");
+
+    add_collectable(tc, worklist, snapshot, tc->instance->env_hash,
+        "Cached environment variable hash");
 }
 
 /* Adds anything that is a root thanks to being referenced by a thread,
@@ -136,24 +148,12 @@ void MVM_gc_root_add_tc_roots_to_worklist(MVMThreadContext *tc, MVMGCWorklist *w
     /* Any exception handler result. */
     add_collectable(tc, worklist, snapshot, tc->last_handler_result, "Last handler result");
 
-    /* The usecapture object. */
-    add_collectable(tc, worklist, snapshot, tc->cur_usecapture, "Cached usecapture");
-
     /* List of SCs currently being compiled. */
     add_collectable(tc, worklist, snapshot, tc->compiling_scs, "Compiling serialization contexts");
 
     /* compunit variable pointer (and be null if thread finished) */
     if (tc->interp_cu)
         add_collectable(tc, worklist, snapshot, *(tc->interp_cu), "Current interpreter compilation unit");
-
-    /* Lexotics cache. */
-    if (tc->lexotic_cache_size) {
-        MVMuint32 i;
-        for (i = 0; i < tc->lexotic_cache_size; i++)
-            if (tc->lexotic_cache[i])
-                add_collectable(tc, worklist, snapshot, tc->lexotic_cache[i], "Lexotic cache entry");
-    }
-
     /* Current dispatcher. */
     add_collectable(tc, worklist, snapshot, tc->cur_dispatcher, "Current dispatcher");
     add_collectable(tc, worklist, snapshot, tc->cur_dispatcher_for, "Current dispatcher for");
@@ -181,6 +181,14 @@ void MVM_gc_root_add_tc_roots_to_worklist(MVMThreadContext *tc, MVMGCWorklist *w
     /* Serialized string heap, if any. */
     add_collectable(tc, worklist, snapshot, tc->serialized_string_heap,
         "Serialized string heap");
+
+    /* Specialization log and stack simulation. */
+    add_collectable(tc, worklist, snapshot, tc->spesh_log, "Specialization log");
+    if (worklist)
+        MVM_spesh_sim_stack_gc_mark(tc, tc->spesh_sim_stack, worklist);
+    else {
+        MVM_spesh_sim_stack_gc_describe(tc, snapshot, tc->spesh_sim_stack);
+    }
 }
 
 /* Pushes a temporary root onto the thread-local roots list. */
@@ -360,23 +368,23 @@ void MVM_gc_root_add_frame_roots_to_worklist(MVMThreadContext *tc, MVMGCWorklist
     MVM_gc_worklist_add(tc, worklist, &cur_frame->code_ref);
     MVM_gc_worklist_add(tc, worklist, &cur_frame->static_info);
 
-    /* Mark special return data, if needed. */
-    if (cur_frame->special_return_data && cur_frame->mark_special_return_data)
-        cur_frame->mark_special_return_data(tc, cur_frame, worklist);
-
-    /* Mark any continuation tags. */
-    if (cur_frame->continuation_tags) {
-        MVMContinuationTag *tag = cur_frame->continuation_tags;
-        while (tag) {
-            MVM_gc_worklist_add(tc, worklist, &tag->tag);
-            tag = tag->next;
+    /* Mark frame extras if needed. */
+    if (cur_frame->extra) {
+        MVMFrameExtra *e = cur_frame->extra;
+        if (e->special_return_data && e->mark_special_return_data)
+            e->mark_special_return_data(tc, cur_frame, worklist);
+        if (e->continuation_tags) {
+            MVMContinuationTag *tag = e->continuation_tags;
+            while (tag) {
+                MVM_gc_worklist_add(tc, worklist, &tag->tag);
+                tag = tag->next;
+            }
         }
+        MVM_gc_worklist_add(tc, worklist, &e->invoked_call_capture);
+        if (e->dynlex_cache_name)
+            MVM_gc_worklist_add(tc, worklist, &e->dynlex_cache_name);
     }
 
-    /* Mark any dyn lex cache. */
-    if (cur_frame->dynlex_cache_name)
-        MVM_gc_worklist_add(tc, worklist, &cur_frame->dynlex_cache_name);
-
     /* Scan the registers. */
     MVM_gc_root_add_frame_registers_to_worklist(tc, worklist, cur_frame);
     scan_lexicals(tc, worklist, cur_frame);
@@ -387,11 +395,16 @@ void MVM_gc_root_add_frame_registers_to_worklist(MVMThreadContext *tc, MVMGCWork
     MVMuint16  i, count, flag;
     MVMuint16 *type_map;
     MVMuint8  *flag_map;
-
     /* We only need to do any of this work if the frame is in dynamic scope. */
     if (frame->work) {
         /* Scan locals. */
-        if (frame->spesh_cand && frame->spesh_log_idx == -1 && frame->spesh_cand->local_types) {
+
+        MVMSpeshCandidate *spesh_cand = frame->spesh_cand;
+        MVMJitCode *jitcode = spesh_cand ? spesh_cand->jitcode : NULL;
+        if (jitcode && jitcode->local_types) {
+            type_map = jitcode->local_types;
+            count    = jitcode->num_locals;
+        } else if (frame->spesh_cand && frame->spesh_cand->local_types) {
             type_map = frame->spesh_cand->local_types;
             count    = frame->spesh_cand->num_locals;
         }
@@ -442,7 +455,7 @@ static void scan_lexicals(MVMThreadContext *tc, MVMGCWorklist *worklist, MVMFram
     if (frame->env) {
         MVMuint16  i, count;
         MVMuint16 *type_map;
-        if (frame->spesh_cand && frame->spesh_log_idx == -1 && frame->spesh_cand->lexical_types) {
+        if (frame->spesh_cand && frame->spesh_cand->lexical_types) {
             type_map = frame->spesh_cand->lexical_types;
             count    = frame->spesh_cand->num_lexicals;
         }
