@@ -32,7 +32,10 @@ typedef struct {
     int fd;
 
     /* Is it seekable? */
-    int seekable;
+    short seekable;
+
+    /* Is it know to be writable? */
+    short known_writable;
 
     /* How many bytes have we read/written? Used to fake tell on handles that
      * are not seekable. */
@@ -40,19 +43,16 @@ typedef struct {
 
     /* Did read already report EOF? */
     int eof_reported;
-} MVMIOFileData;
 
-/* Closes the file. */
-static MVMint64 closefh(MVMThreadContext *tc, MVMOSHandle *h) {
-    MVMIOFileData *data = (MVMIOFileData *)h->body.data;
-    if (data->fd != -1) {
-        int r = close(data->fd);
-        data->fd = -1;
-        if (r == -1)
-            MVM_exception_throw_adhoc(tc, "Failed to close filehandle: %s", strerror(errno));
-    }
-    return 0;
-}
+    /* Output buffer, for buffered output. */
+    char *output_buffer;
+
+    /* Size of the output buffer, for buffered output; 0 if not buffering. */
+    size_t output_buffer_size;
+
+    /* How much of the output buffer has been used so far. */
+    size_t output_buffer_used;
+} MVMIOFileData;
 
 /* Checks if the file is a TTY. */
 static MVMint64 is_tty(MVMThreadContext *tc, MVMOSHandle *h) {
@@ -66,18 +66,53 @@ static MVMint64 mvm_fileno(MVMThreadContext *tc, MVMOSHandle *h) {
     return (MVMint64)data->fd;
 }
 
+/* Performs a write, either because a buffer filled or because we are not
+ * buffering output. */
+static void perform_write(MVMThreadContext *tc, MVMIOFileData *data, char *buf, MVMint64 bytes) {
+    MVMint64 bytes_written = 0;
+    MVM_gc_mark_thread_blocked(tc);
+    while (bytes > 0) {
+        int r;
+        do {
+            r = write(data->fd, buf, (int)bytes);
+        } while (r == -1 && errno == EINTR);
+        if (r == -1) {
+            int save_errno = errno;
+            MVM_gc_mark_thread_unblocked(tc);
+            MVM_exception_throw_adhoc(tc, "Failed to write bytes to filehandle: %s",
+                strerror(save_errno));
+        }
+        bytes_written += r;
+        buf += r;
+        bytes -= r;
+    }
+    MVM_gc_mark_thread_unblocked(tc);
+    data->byte_position += bytes_written;
+    data->known_writable = 1;
+}
+
+/* Flushes any existing output buffer and clears use back to 0. */
+static void flush_output_buffer(MVMThreadContext *tc, MVMIOFileData *data) {
+    if (data->output_buffer_used) {
+        perform_write(tc, data, data->output_buffer, data->output_buffer_used);
+        data->output_buffer_used = 0;
+    }
+}
+
 /* Seek to the specified position in the file. */
 static void seek(MVMThreadContext *tc, MVMOSHandle *h, MVMint64 offset, MVMint64 whence) {
     MVMIOFileData *data = (MVMIOFileData *)h->body.data;
     if (!data->seekable)
         MVM_exception_throw_adhoc(tc, "It is not possible to seek this kind of handle");
+    flush_output_buffer(tc, data);
     if (MVM_platform_lseek(data->fd, offset, whence) == -1)
         MVM_exception_throw_adhoc(tc, "Failed to seek in filehandle: %d", errno);
 }
 
-/* Get curernt position in the file. */
+/* Get current position in the file. */
 static MVMint64 mvm_tell(MVMThreadContext *tc, MVMOSHandle *h) {
     MVMIOFileData *data = (MVMIOFileData *)h->body.data;
+    flush_output_buffer(tc, data);
     if (data->seekable) {
         MVMint64 r;
         if ((r = MVM_platform_lseek(data->fd, 0, SEEK_CUR)) == -1)
@@ -103,16 +138,19 @@ static MVMint64 read_bytes(MVMThreadContext *tc, MVMOSHandle *h, char **buf_out,
     if (bytes > 16387 && _isatty(data->fd))
         bytes = 16387;
 #endif
-    MVM_gc_mark_thread_blocked(tc);
-    if ((bytes_read = read(data->fd, buf, bytes)) == -1) {
+    flush_output_buffer(tc, data);
+    do {
+        MVM_gc_mark_thread_blocked(tc);
+        bytes_read = read(data->fd, buf, bytes);
+        MVM_gc_mark_thread_unblocked(tc);
+    } while(bytes_read == -1 && errno == EINTR);
+    if (bytes_read  == -1) {
         int save_errno = errno;
         MVM_free(buf);
-        MVM_gc_mark_thread_unblocked(tc);
         MVM_exception_throw_adhoc(tc, "Reading from filehandle failed: %s",
             strerror(save_errno));
     }
     *buf_out = buf;
-    MVM_gc_mark_thread_unblocked(tc);
     MVM_telemetry_interval_annotate(bytes_read, interval_id, "read this many bytes");
     MVM_telemetry_interval_stop(tc, interval_id, "syncfile.read_to_buffer");
     data->byte_position += bytes_read;
@@ -135,42 +173,69 @@ static MVMint64 mvm_eof(MVMThreadContext *tc, MVMOSHandle *h) {
         /* Comparison with seek_pos for some special files, like those in /proc,
          * which file size is 0 can be false. In that case, we fall back to check
          * file size to detect EOF. */
-        return statbuf.st_size == seek_pos || statbuf.st_size == 0;
+        return statbuf.st_size <= seek_pos || statbuf.st_size == 0;
     }
     else {
         return data->eof_reported;
     }
 }
 
+/* Sets the output buffer size; if <= 0, means no buffering. Flushes any
+ * existing buffer before changing. */
+static void set_buffer_size(MVMThreadContext *tc, MVMOSHandle *h, MVMint64 size) {
+    MVMIOFileData *data = (MVMIOFileData *)h->body.data;
+
+    /* Flush and clear up any existing output buffer. */
+    flush_output_buffer(tc, data);
+    MVM_free(data->output_buffer);
+
+    /* Set up new buffer if needed. */
+    if (size > 0) {
+        data->output_buffer_size = size;
+        data->output_buffer = MVM_malloc(size);
+    }
+    else {
+        data->output_buffer_size = 0;
+        data->output_buffer = NULL;
+    }
+}
+
 /* Writes the specified bytes to the file handle. */
 static MVMint64 write_bytes(MVMThreadContext *tc, MVMOSHandle *h, char *buf, MVMint64 bytes) {
     MVMIOFileData *data = (MVMIOFileData *)h->body.data;
-    MVMint64 bytes_written = 0;
-    MVM_gc_mark_thread_blocked(tc);
-    while (bytes > 0) {
-        int r = write(data->fd, buf, (int)bytes);
-        if (r == -1) {
-            int save_errno = errno;
-            MVM_gc_mark_thread_unblocked(tc);
-            MVM_exception_throw_adhoc(tc, "Failed to write bytes to filehandle: %s",
-                strerror(save_errno));
+    if (data->output_buffer_size && data->known_writable) {
+        /* If we can't fit it on the end of the buffer, flush the buffer. */
+        if (data->output_buffer_used + bytes > data->output_buffer_size)
+            flush_output_buffer(tc, data);
+
+        /* If we can fit it in the buffer now, memcpy it there, and we're
+         * done. */
+        if (bytes < data->output_buffer_size) {
+            memcpy(data->output_buffer + data->output_buffer_used, buf, bytes);
+            data->output_buffer_used += bytes;
+            return bytes;
         }
-        bytes_written += r;
-        buf += r;
-        bytes -= r;
     }
-    MVM_gc_mark_thread_unblocked(tc);
-    data->byte_position += bytes_written;
-    return bytes_written;
+    perform_write(tc, data, buf, bytes);
+    return bytes;
 }
 
 /* Flushes the file handle. */
-static void flush(MVMThreadContext *tc, MVMOSHandle *h){
+static void flush(MVMThreadContext *tc, MVMOSHandle *h, MVMint32 sync){
     MVMIOFileData *data = (MVMIOFileData *)h->body.data;
-    if (MVM_platform_fsync(data->fd) == -1) {
-        /* If this is something that can't be flushed, we let that pass. */
-        if (errno != EROFS && errno != EINVAL)
-            MVM_exception_throw_adhoc(tc, "Failed to flush filehandle: %s", strerror(errno));
+    flush_output_buffer(tc, data);
+    if (sync) {
+        if (MVM_platform_fsync(data->fd) == -1) {
+            /* If this is something that can't be flushed, we let that pass. */
+            if (errno != EROFS && errno != EINVAL
+#ifdef WSL_BASH_ON_WIN
+               && ! (errno == EIO && ! data->seekable)
+               /* Bash on Win10 doesn't seem to handle TTYs right when flushing:
+                * https://github.com/borgbackup/borg/issues/1961#issuecomment-335653560 */
+#endif
+            )
+                MVM_exception_throw_adhoc(tc, "Failed to flush filehandle: %s", strerror(errno));
+        }
     }
 }
 
@@ -181,11 +246,20 @@ static void truncatefh(MVMThreadContext *tc, MVMOSHandle *h, MVMint64 bytes) {
         MVM_exception_throw_adhoc(tc, "Failed to truncate filehandle: %s", strerror(errno));
 }
 
-/* Operations aiding process spawning and I/O handling. */
-static void bind_stdio_handle(MVMThreadContext *tc, MVMOSHandle *h, uv_stdio_container_t *stdio) {
+/* Closes the file. */
+static MVMint64 closefh(MVMThreadContext *tc, MVMOSHandle *h) {
     MVMIOFileData *data = (MVMIOFileData *)h->body.data;
-    stdio->flags        = UV_INHERIT_FD;
-    stdio->data.fd      = data->fd;
+    if (data->fd != -1) {
+        int r;
+        flush_output_buffer(tc, data);
+        MVM_free(data->output_buffer);
+        data->output_buffer = NULL;
+        r = close(data->fd);
+        data->fd = -1;
+        if (r == -1)
+            MVM_exception_throw_adhoc(tc, "Failed to close filehandle: %s", strerror(errno));
+    }
+    return 0;
 }
 
 /* Locks a file. */
@@ -207,9 +281,12 @@ static MVMint64 lock(MVMThreadContext *tc, MVMOSHandle *h, MVMint64 flag) {
                                        ? 0 : LOCKFILE_EXCLUSIVE_LOCK);
 
     memset (&offset, 0, sizeof(offset));
+    MVM_gc_mark_thread_blocked(tc);
     if (LockFileEx(hf, flag, 0, len, len, &offset)) {
+        MVM_gc_mark_thread_unblocked(tc);
         return 1;
     }
+    MVM_gc_mark_thread_unblocked(tc);
 
     MVM_exception_throw_adhoc(tc, "Failed to lock filehandle: %d", GetLastError());
 
@@ -234,7 +311,9 @@ static MVMint64 lock(MVMThreadContext *tc, MVMOSHandle *h, MVMint64 flag) {
     fc = (flag & MVM_FILE_FLOCK_NONBLOCK) ? F_SETLK : F_SETLKW;
 
     do {
+        MVM_gc_mark_thread_blocked(tc);
         r = fcntl(fd, fc, &l);
+        MVM_gc_mark_thread_unblocked(tc);
     } while (r == -1 && errno == EINTR);
 
     if (r == -1) {
@@ -260,9 +339,12 @@ static void unlock(MVMThreadContext *tc, MVMOSHandle *h) {
     }
 
     memset (&offset, 0, sizeof(offset));
+    MVM_gc_mark_thread_blocked(tc);
     if (UnlockFileEx(hf, 0, len, len, &offset)) {
+        MVM_gc_mark_thread_unblocked(tc);
         return;
     }
+    MVM_gc_mark_thread_unblocked(tc);
 
     MVM_exception_throw_adhoc(tc, "Failed to unlock filehandle: %d", GetLastError());
 #else
@@ -277,7 +359,9 @@ static void unlock(MVMThreadContext *tc, MVMOSHandle *h) {
     l.l_type = F_UNLCK;
 
     do {
+        MVM_gc_mark_thread_blocked(tc);
         r = fcntl(fd, F_SETLKW, &l);
+        MVM_gc_mark_thread_unblocked(tc);
     } while (r == -1 && errno == EINTR);
 
     if (r == -1) {
@@ -289,8 +373,10 @@ static void unlock(MVMThreadContext *tc, MVMOSHandle *h) {
 /* Frees data associated with the handle. */
 static void gc_free(MVMThreadContext *tc, MVMObject *h, void *d) {
     MVMIOFileData *data = (MVMIOFileData *)d;
-    if (data)
+    if (data) {
+        MVM_free(data->output_buffer);
         MVM_free(data);
+    }
 }
 
 /* IO ops table, populated with functions. */
@@ -298,7 +384,6 @@ static const MVMIOClosable      closable      = { closefh };
 static const MVMIOSyncReadable  sync_readable = { read_bytes, mvm_eof };
 static const MVMIOSyncWritable  sync_writable = { write_bytes, flush, truncatefh };
 static const MVMIOSeekable      seekable      = { seek, mvm_tell };
-static const MVMIOPipeable      pipeable      = { bind_stdio_handle };
 static const MVMIOLockable      lockable      = { lock, unlock };
 static const MVMIOIntrospection introspection = { is_tty, mvm_fileno };
 
@@ -311,9 +396,10 @@ static const MVMIOOps op_table = {
     NULL,
     &seekable,
     NULL,
-    &pipeable,
+    NULL,
     &lockable,
     &introspection,
+    &set_buffer_size,
     NULL,
     gc_free
 };
@@ -397,7 +483,7 @@ MVMObject * MVM_file_open_fh(MVMThreadContext *tc, MVMString *filename, MVMStrin
         MVMOSHandle   * const result = (MVMOSHandle *)MVM_repr_alloc_init(tc,
             tc->instance->boot_types.BOOTIO);
         data->fd          = fd;
-        data->seekable    = MVM_platform_lseek(fd, 0, SEEK_CUR) != -1;
+        data->seekable    = MVM_platform_is_fd_seekable(fd);
         result->body.ops  = &op_table;
         result->body.data = data;
         return (MVMObject *)result;
@@ -409,7 +495,7 @@ MVMObject * MVM_file_handle_from_fd(MVMThreadContext *tc, int fd) {
     MVMOSHandle   * const result = (MVMOSHandle *)MVM_repr_alloc_init(tc, tc->instance->boot_types.BOOTIO);
     MVMIOFileData * const data   = MVM_calloc(1, sizeof(MVMIOFileData));
     data->fd          = fd;
-    data->seekable    = MVM_platform_lseek(fd, 0, SEEK_CUR) != -1;
+    data->seekable    = MVM_platform_is_fd_seekable(fd);
     result->body.ops  = &op_table;
     result->body.data = data;
 #ifdef _WIN32
