@@ -2,10 +2,14 @@
 
 #ifdef _WIN32
     #include <winsock2.h>
+    #include <ws2tcpip.h>
     typedef SOCKET Socket;
+    #define sa_family_t unsigned int
 #else
     #include "unistd.h"
     #include <sys/socket.h>
+    #include <sys/un.h>
+
     typedef int Socket;
     #define closesocket close
 #endif
@@ -14,12 +18,13 @@
 #define snprintf _snprintf
 #endif
 
-/* Assuemd maximum packet size. If ever changing this to something beyond a
+/* Assumed maximum packet size. If ever changing this to something beyond a
  * 16-bit number, then make sure to change the receive offsets in the data
  * structure below. */
 #define PACKET_SIZE 65535
 
 /* Error handling varies between POSIX and WinSock. */
+MVM_NO_RETURN static void throw_error(MVMThreadContext *tc, int r, char *operation) MVM_NO_RETURN_GCC;
 #ifdef _WIN32
     #define MVM_IS_SOCKET_ERROR(x) ((x) == SOCKET_ERROR)
     static void throw_error(MVMThreadContext *tc, int r, char *operation) {
@@ -61,10 +66,12 @@ typedef struct {
 static void read_one_packet(MVMThreadContext *tc, MVMIOSyncSocketData *data) {
     unsigned int interval_id = MVM_telemetry_interval_start(tc, "syncsocket.read_one_packet");
     int r;
-    MVM_gc_mark_thread_blocked(tc);
     data->last_packet = MVM_malloc(PACKET_SIZE);
-    r = recv(data->handle, data->last_packet, PACKET_SIZE, 0);
-    MVM_gc_mark_thread_unblocked(tc);
+    do {
+        MVM_gc_mark_thread_blocked(tc);
+        r = recv(data->handle, data->last_packet, PACKET_SIZE, 0);
+        MVM_gc_mark_thread_unblocked(tc);
+    } while(r == -1 && errno == EINTR);
     MVM_telemetry_interval_stop(tc, interval_id, "syncsocket.read_one_packet");
     if (MVM_IS_SOCKET_ERROR(r) || r == 0) {
         MVM_free(data->last_packet);
@@ -178,7 +185,7 @@ MVMint64 socket_eof(MVMThreadContext *tc, MVMOSHandle *h) {
     return data->eof;
 }
 
-void socket_flush(MVMThreadContext *tc, MVMOSHandle *h) {
+void socket_flush(MVMThreadContext *tc, MVMOSHandle *h, MVMint32 sync) {
     /* A no-op for sockets; we don't buffer. */
 }
 
@@ -195,7 +202,10 @@ MVMint64 socket_write_bytes(MVMThreadContext *tc, MVMOSHandle *h, char *buf, MVM
     interval_id = MVM_telemetry_interval_start(tc, "syncsocket.write_bytes");
     MVM_gc_mark_thread_blocked(tc);
     while (bytes > 0) {
-        int r = send(data->handle, buf, (int)bytes, 0);
+        int r;
+        do {
+            r = send(data->handle, buf, (int)bytes, 0);
+        } while(r == -1 && errno == EINTR);
         if (MVM_IS_SOCKET_ERROR(r)) {
             MVM_gc_mark_thread_unblocked(tc);
             MVM_telemetry_interval_stop(tc, interval_id, "syncsocket.write_bytes");
@@ -228,30 +238,92 @@ static void gc_free(MVMThreadContext *tc, MVMObject *h, void *d) {
     MVM_free(data);
 }
 
-/* Actually, it may return sockaddr_in6 as well; it's not a problem for us, because we just
- * pass is straight to uv, and the first thing it does is it looks at the address family,
- * but it's a thing to remember if someone feels like peeking inside the returned struct. */
+static size_t get_struct_size_for_family(sa_family_t family) {
+    switch (family) {
+        case AF_INET6:
+            return sizeof(struct sockaddr_in6);
+        case AF_INET:
+            return sizeof(struct sockaddr_in);
+#ifndef _WIN32
+        case AF_UNIX:
+            return sizeof(struct sockaddr_un);
+#endif
+        default:
+            return sizeof(struct sockaddr);
+    }
+}
+
+/* This function may return any type of sockaddr e.g. sockaddr_un, sockaddr_in or sockaddr_in6
+ * It shouldn't be a problem with general code as long as the port number is kept below the int16 limit: 65536
+ * After this it defines the family which may spawn non internet sockaddr's
+ * The family can be extracted by (port >> 16) & USHORT_MAX
+ *
+ * Currently supported families:
+ *
+ * AF_UNSPEC = 1
+ *   Unspecified, in most cases should be equal to AF_INET or AF_INET6
+ *
+ * AF_UNIX = 1
+ *   Unix domain socket, will spawn a sockaddr_un which will use the given host as path
+ *   e.g: MVM_io_resolve_host_name(tc, "/run/moarvm.sock", 1 << 16)
+ *   will spawn an unix domain socket on /run/moarvm.sock
+ *
+ * AF_INET = 2
+ *   IPv4 socket
+ *
+ * AF_INET6 = 10
+ *   IPv6 socket
+ */
 struct sockaddr * MVM_io_resolve_host_name(MVMThreadContext *tc, MVMString *host, MVMint64 port) {
     char *host_cstr = MVM_string_utf8_encode_C_string(tc, host);
     struct sockaddr *dest;
-    struct addrinfo *result;
     int error;
+    struct addrinfo *result;
     char port_cstr[8];
+    unsigned short family = (port >> 16) & USHRT_MAX;
+    struct addrinfo hints;
+
+#ifndef _WIN32
+    /* AF_UNIX = 1 */
+    if (family == AF_UNIX) {
+        struct sockaddr_un *result_un = MVM_malloc(sizeof(struct sockaddr_un));
+
+        if (strlen(host_cstr) > 107) {
+            MVM_free(result_un);
+            MVM_free(host_cstr);
+            MVM_exception_throw_adhoc(tc, "Socket path can only be maximal 107 characters long");
+        }
+
+        result_un->sun_family = AF_UNIX;
+        strcpy(result_un->sun_path, host_cstr);
+        MVM_free(host_cstr);
+
+        return (struct sockaddr *)result_un;
+    }
+#endif
+
+    hints.ai_family = family;
+    hints.ai_socktype = 0;
+    hints.ai_flags = AI_PASSIVE;
+    hints.ai_protocol = 0;
+    hints.ai_addrlen = 0;
+    hints.ai_addr = NULL;
+    hints.ai_canonname = NULL;
+    hints.ai_next = NULL;
+
     snprintf(port_cstr, 8, "%d", (int)port);
 
-    error = getaddrinfo(host_cstr, port_cstr, NULL, &result);
-    MVM_free(host_cstr);
+    error = getaddrinfo(host_cstr, port_cstr, &hints, &result);
     if (error == 0) {
-        if (result->ai_addr->sa_family == AF_INET6) {
-            dest = MVM_malloc(sizeof(struct sockaddr_in6));
-            memcpy(dest, result->ai_addr, sizeof(struct sockaddr_in6));
-        } else {
-            dest = MVM_malloc(sizeof(struct sockaddr));
-            memcpy(dest, result->ai_addr, sizeof(struct sockaddr));
-        }
+        size_t size = get_struct_size_for_family(result->ai_addr->sa_family);
+        MVM_free(host_cstr);
+        dest = MVM_malloc(size);
+        memcpy(dest, result->ai_addr, size);
     }
     else {
-        MVM_exception_throw_adhoc(tc, "Failed to resolve host name");
+        char *waste[] = { host_cstr, NULL };
+        MVM_exception_throw_adhoc_free(tc, waste, "Failed to resolve host name '%s' with family %d. Error: '%s'",
+                                       host_cstr, family, gai_strerror(error));
     }
     freeaddrinfo(result);
 
@@ -275,11 +347,11 @@ static void socket_connect(MVMThreadContext *tc, MVMOSHandle *h, MVMString *host
             throw_error(tc, s, "create socket");
         }
 
-        MVM_gc_mark_thread_blocked(tc);
-        r = connect(s, dest, dest->sa_family == AF_INET6
-                ? sizeof(struct sockaddr_in6)
-                : sizeof(struct sockaddr));
-        MVM_gc_mark_thread_unblocked(tc);
+        do {
+            MVM_gc_mark_thread_blocked(tc);
+            r = connect(s, dest, (socklen_t)get_struct_size_for_family(dest->sa_family));
+            MVM_gc_mark_thread_unblocked(tc);
+        } while(r == -1 && errno == EINTR);
         MVM_free(dest);
         if (MVM_IS_SOCKET_ERROR(r)) {
             MVM_telemetry_interval_stop(tc, interval_id, "syncsocket connect");
@@ -319,9 +391,7 @@ static void socket_bind(MVMThreadContext *tc, MVMOSHandle *h, MVMString *host, M
         }
 #endif
 
-        r = bind(s, dest, dest->sa_family == AF_INET6
-                ? sizeof(struct sockaddr_in6)
-                : sizeof(struct sockaddr));
+        r = bind(s, dest, (socklen_t)get_struct_size_for_family(dest->sa_family));
         MVM_free(dest);
         if (MVM_IS_SOCKET_ERROR(r))
             throw_error(tc, s, "bind socket");
@@ -341,7 +411,8 @@ MVMint64 socket_getport(MVMThreadContext *tc, MVMOSHandle *h) {
     MVMIOSyncSocketData *data = (MVMIOSyncSocketData *)h->body.data;
 
     struct sockaddr_storage name;
-    int error, len = sizeof(struct sockaddr_storage);
+    int error;
+    socklen_t len = sizeof(struct sockaddr_storage);
     MVMint64 port = 0;
 
     error = getsockname(data->handle, (struct sockaddr *) &name, &len);
@@ -387,6 +458,7 @@ static const MVMIOOps op_table = {
     NULL,
     NULL,
     NULL,
+    NULL,
     gc_free
 };
 
@@ -395,9 +467,11 @@ static MVMObject * socket_accept(MVMThreadContext *tc, MVMOSHandle *h) {
     Socket s;
 
     unsigned int interval_id = MVM_telemetry_interval_start(tc, "syncsocket accept");
-    MVM_gc_mark_thread_blocked(tc);
-    s = accept(data->handle, NULL, NULL);
-    MVM_gc_mark_thread_unblocked(tc);
+    do {
+        MVM_gc_mark_thread_blocked(tc);
+        s = accept(data->handle, NULL, NULL);
+        MVM_gc_mark_thread_unblocked(tc);
+    } while(s == -1 && errno == EINTR);
     if (MVM_IS_SOCKET_ERROR(s)) {
         MVM_telemetry_interval_stop(tc, interval_id, "syncsocket accept failed");
         throw_error(tc, s, "accept socket connection");
