@@ -47,32 +47,57 @@ static void demand_extop(MVMThreadContext *tc, MVMCompUnit *target_cu, MVMCompUn
  * already identify a spesh candidate. Returns NULL if no inlining is possible
  * or a graph ready to be merged if it will be possible. */
 MVMSpeshGraph * MVM_spesh_inline_try_get_graph(MVMThreadContext *tc, MVMSpeshGraph *inliner,
-                                               MVMCode *target, MVMSpeshCandidate *cand) {
+                                               MVMStaticFrame *target_sf,
+                                               MVMSpeshCandidate *cand,
+                                               MVMSpeshIns *invoke_ins,
+                                               char **no_inline_reason) {
     MVMSpeshGraph *ig;
-    MVMSpeshBB    *bb;
+    MVMSpeshBB *bb;
+    MVMint32 same_hll;
 
     /* Check inlining is enabled. */
-    if (!tc->instance->spesh_inline_enabled)
+    if (!tc->instance->spesh_inline_enabled) {
+        *no_inline_reason = "inlining is disabled";
         return NULL;
+    }
+
+    /* Check frame is not marked as not being allowed to be inlined. */
+    if (target_sf->body.no_inline) {
+        *no_inline_reason = "the frame is marked as no-inline";
+        return NULL;
+    }
 
     /* Check bytecode size is within the inline limit. */
-    if (cand->bytecode_size > MVM_SPESH_MAX_INLINE_SIZE)
+    if (cand->bytecode_size > MVM_SPESH_MAX_INLINE_SIZE) {
+        *no_inline_reason = "bytecode is too large to inline";
         return NULL;
+    }
 
     /* Ensure that this isn't a recursive inlining. */
-    if (target->body.sf == inliner->sf)
+    if (target_sf == inliner->sf) {
+        *no_inline_reason = "recursive calls cannot be inlined";
         return NULL;
+    }
 
-    /* Ensure they're from the same HLL. */
-    if (target->body.sf->body.cu->body.hll_config != inliner->sf->body.cu->body.hll_config)
+    /* Ensure it has no state vars (these need the setup code in frame
+     * invoke). */
+    if (target_sf->body.has_state_vars) {
+        *no_inline_reason = "cannot inline code that declares a state variable";
         return NULL;
+    }
 
-    /* Ensure the candidate isn't still logging. */
-    if (cand->sg)
+    /* Ensure it's not a thunk (need to skip over those in exception search). */
+    if (target_sf->body.is_thunk) {
+        *no_inline_reason = "cannot inline code marked as a thunk";
         return NULL;
+    }
+
+    /* If they're from the same HLL, we'll need to watch out for ops that are
+     * HLL sensitive. */
+    same_hll = target_sf->body.cu->body.hll_config == inliner->sf->body.cu->body.hll_config;
 
     /* Build graph from the already-specialized bytecode. */
-    ig = MVM_spesh_graph_create_from_cand(tc, target->body.sf, cand, 0);
+    ig = MVM_spesh_graph_create_from_cand(tc, target_sf, cand, 0);
 
     /* Traverse graph, looking for anything that might prevent inlining and
      * also building usage counts up. */
@@ -94,32 +119,53 @@ MVMSpeshGraph * MVM_spesh_inline_try_get_graph(MVMThreadContext *tc, MVMSpeshGra
 
             /* Instruction may be marked directly as not being inlinable, in
              * which case we're done. */
-            if (!is_phi && ins->info->no_inline)
+            if (!is_phi && ins->info->no_inline) {
+                *no_inline_reason = "target has a :noinline instruction";
                 goto not_inlinable;
-
-            /* If we have lexical access, make sure it's within the frame. */
-            if (ins->info->opcode == MVM_OP_getlex) {
-                if (ins->operands[1].lex.outers > 0)
-                    goto not_inlinable;
             }
-            else if (ins->info->opcode == MVM_OP_bindlex) {
-                if (ins->operands[0].lex.outers > 0)
+
+            /* If we don't have the same HLL and there's a :useshll op, we
+             * cannot inline. */
+            if (!same_hll && ins->info->uses_hll) {
+                *no_inline_reason = "target has a :useshll instruction and HLLs are different";
+                goto not_inlinable;
+            }
+
+            /* If we have an invoke_o, but a return_[ins] that would require
+             * boxing, we can't inline if it's not the same HLL. */
+            if (!same_hll && invoke_ins->info->opcode == MVM_OP_invoke_o) {
+                switch (ins->info->opcode) {
+                    case MVM_OP_return_i:
+                    case MVM_OP_return_n:
+                    case MVM_OP_return_s:
+                        *no_inline_reason = "target needs a return boxing and HLLs are different";
+                        goto not_inlinable;
+                }
+            }
+
+            /* If we have lexical bind, make sure it's within the frame. */
+            if (ins->info->opcode == MVM_OP_bindlex) {
+                if (ins->operands[0].lex.outers > 0) {
+                    *no_inline_reason = "target has bind to outer lexical";
                     goto not_inlinable;
+                }
             }
 
             /* Check we don't have too many args for inlining to work out. */
-            if (ins->info->opcode == MVM_OP_sp_getarg_o ||
+            else if (ins->info->opcode == MVM_OP_sp_getarg_o ||
                     ins->info->opcode == MVM_OP_sp_getarg_i ||
                     ins->info->opcode == MVM_OP_sp_getarg_n ||
                     ins->info->opcode == MVM_OP_sp_getarg_s) {
-                if (ins->operands[1].lit_i16 >= MAX_ARGS_FOR_OPT)
+                if (ins->operands[1].lit_i16 >= MAX_ARGS_FOR_OPT) {
+                    *no_inline_reason = "too many arguments to inline";
                     goto not_inlinable;
+                }
             }
 
             /* Ext-ops need special care in inter-comp-unit inlines. */
             if (ins->info->opcode == (MVMuint16)-1) {
                 MVMCompUnit *target_cu = inliner->sf->body.cu;
-                MVMCompUnit *source_cu = target->body.sf->body.cu;
+                MVMCompUnit *source_cu = target_sf->body.cu;
                 if (source_cu != target_cu)
                     demand_extop(tc, target_cu, source_cu, ins->info);
             }
@@ -199,8 +245,9 @@ static void resize_handlers_table(MVMThreadContext *tc, MVMSpeshGraph *inliner,
     if (inliner->handlers == inliner->sf->body.handlers) {
         /* Original handlers table; need a copy. */
         MVMFrameHandler *new_handlers = MVM_malloc(new_handler_count * sizeof(MVMFrameHandler));
-        memcpy(new_handlers, inliner->handlers,
-            inliner->num_handlers * sizeof(MVMFrameHandler));
+        if (inliner->handlers)
+            memcpy(new_handlers, inliner->handlers,
+                inliner->num_handlers * sizeof(MVMFrameHandler));
         inliner->handlers = new_handlers;
     }
     else {
@@ -210,13 +257,29 @@ static void resize_handlers_table(MVMThreadContext *tc, MVMSpeshGraph *inliner,
     }
 }
 
+/* Rewrites a lexical lookup to an outer to be done via. a register holding
+ * the outer coderef. */
+static void rewrite_outer_lookup(MVMThreadContext *tc, MVMSpeshGraph *g,
+                                 MVMSpeshIns *ins, MVMuint16 num_locals,
+                                 MVMuint16 op, MVMSpeshOperand code_ref_reg) {
+    MVMSpeshOperand *new_operands = MVM_spesh_alloc(tc, g, 4 * sizeof(MVMSpeshOperand));
+    new_operands[0] = ins->operands[0];
+    new_operands[0].reg.orig += num_locals;
+    new_operands[1].lit_ui16 = ins->operands[1].lex.idx;
+    new_operands[2].lit_ui16 = ins->operands[1].lex.outers;
+    new_operands[3] = code_ref_reg;
+    ins->info = MVM_op_get_op(op);
+    ins->operands = new_operands;
+}
+
 /* Merges the inlinee's spesh graph into the inliner. */
 static void merge_graph(MVMThreadContext *tc, MVMSpeshGraph *inliner,
-                 MVMSpeshGraph *inlinee, MVMCode *inlinee_code,
-                 MVMSpeshIns *invoke_ins) {
+                 MVMSpeshGraph *inlinee, MVMStaticFrame *inlinee_sf,
+                 MVMSpeshIns *invoke_ins, MVMSpeshOperand code_ref_reg,
+                 MVMuint32 *inline_boundary_handler) {
     MVMSpeshFacts **merged_facts;
     MVMuint16      *merged_fact_counts;
-    MVMint32        i, total_inlines, orig_deopt_addrs;
+    MVMint32        i, orig_inlines, total_inlines, orig_deopt_addrs;
     MVMSpeshBB     *inlinee_first_bb = NULL, *inlinee_last_bb = NULL;
     MVMint32        active_handlers_at_invoke = 0;
 
@@ -254,6 +317,26 @@ static void merge_graph(MVMThreadContext *tc, MVMSpeshGraph *inliner,
                 for (i = 0; i < ins->info->num_operands; i++)
                     ins->operands[i].reg.orig += inliner->num_locals;
             }
+            else if (opcode == MVM_OP_sp_getlex_o && ins->operands[1].lex.outers > 0) {
+                rewrite_outer_lookup(tc, inliner, ins, inliner->num_locals,
+                    MVM_OP_sp_getlexvia_o, code_ref_reg);
+            }
+            else if (opcode == MVM_OP_sp_getlex_ins && ins->operands[1].lex.outers > 0) {
+                rewrite_outer_lookup(tc, inliner, ins, inliner->num_locals,
+                    MVM_OP_sp_getlexvia_ins, code_ref_reg);
+            }
+            else if (opcode == MVM_OP_getlex && ins->operands[1].lex.outers > 0) {
+                MVMuint16 outers = ins->operands[1].lex.outers;
+                MVMStaticFrame *outer = inlinee_sf;
+                while (outers--)
+                    outer = outer->body.outer;
+                if (outer->body.lexical_types[ins->operands[1].lex.idx] == MVM_reg_obj)
+                    rewrite_outer_lookup(tc, inliner, ins, inliner->num_locals,
+                        MVM_OP_sp_getlexvia_o, code_ref_reg);
+                else
+                    rewrite_outer_lookup(tc, inliner, ins, inliner->num_locals,
+                        MVM_OP_sp_getlexvia_ins, code_ref_reg);
+            }
             else {
                 for (i = 0; i < ins->info->num_operands; i++) {
                     MVMuint8 flags = ins->info->operands[i];
@@ -289,12 +372,6 @@ static void merge_graph(MVMThreadContext *tc, MVMSpeshGraph *inliner,
                 }
             }
 
-            /* Since inlining eliminates the caller/callee distinction, we
-             * need skip going up a caller when resolving exceptions in a
-             * caller-relative way. */
-            if (ins->info->opcode == MVM_OP_throwpayloadlexcaller)
-                ins->info = MVM_op_get_op(MVM_OP_throwpayloadlex);
-
             ins = ins->next;
         }
         bb->idx += inliner->num_bbs - 1; /* -1 as we won't include entry */
@@ -318,6 +395,17 @@ static void merge_graph(MVMThreadContext *tc, MVMSpeshGraph *inliner,
         }
     }
 
+    /* Make all of the inlinee's entry block's successors (except the linear
+     * next) also be successors of the inliner's entry block; this keeps any
+     * exception handlers alive in the graph. */
+    while (inlinee->entry->num_succ > 1) {
+        MVMSpeshBB *move = inlinee->entry->succ[0] == inlinee->entry->linear_next
+            ? inlinee->entry->succ[1]
+            : inlinee->entry->succ[0];
+        MVM_spesh_manipulate_remove_successor(tc, inlinee->entry, move);
+        MVM_spesh_manipulate_add_successor(tc, inliner, inliner->entry, move);
+    }
+
     /* Merge facts. */
     merged_facts = MVM_spesh_alloc(tc, inliner,
         (inliner->num_locals + inlinee->num_locals) * sizeof(MVMSpeshFacts *));
@@ -373,6 +461,7 @@ static void merge_graph(MVMThreadContext *tc, MVMSpeshGraph *inliner,
     }
 
     /* Merge inlines table, and add us an entry too. */
+    orig_inlines = inliner->num_inlines;
     total_inlines = inliner->num_inlines + inlinee->num_inlines + 1;
     inliner->inlines = inliner->num_inlines
         ? MVM_realloc(inliner->inlines, total_inlines * sizeof(MVMSpeshInline))
@@ -381,14 +470,17 @@ static void merge_graph(MVMThreadContext *tc, MVMSpeshGraph *inliner,
         memcpy(inliner->inlines + inliner->num_inlines, inlinee->inlines,
             inlinee->num_inlines * sizeof(MVMSpeshInline));
     for (i = inliner->num_inlines; i < total_inlines - 1; i++) {
+        inliner->inlines[i].code_ref_reg += inliner->num_locals;
         inliner->inlines[i].locals_start += inliner->num_locals;
         inliner->inlines[i].lexicals_start += inliner->num_lexicals;
         inliner->inlines[i].return_deopt_idx += orig_deopt_addrs;
     }
-    inliner->inlines[total_inlines - 1].code           = inlinee_code;
+    inliner->inlines[total_inlines - 1].sf             = inlinee_sf;
+    inliner->inlines[total_inlines - 1].code_ref_reg   = code_ref_reg.reg.orig;
     inliner->inlines[total_inlines - 1].g              = inlinee;
     inliner->inlines[total_inlines - 1].locals_start   = inliner->num_locals;
     inliner->inlines[total_inlines - 1].lexicals_start = inliner->num_lexicals;
+    inliner->inlines[total_inlines - 1].num_locals     = inlinee->num_locals;
     switch (invoke_ins->info->opcode) {
     case MVM_OP_invoke_v:
         inliner->inlines[total_inlines - 1].res_type = MVM_RETURN_VOID;
@@ -413,6 +505,9 @@ static void merge_graph(MVMThreadContext *tc, MVMSpeshGraph *inliner,
         MVM_oops(tc, "Spesh inline: unknown invoke instruction");
     }
     inliner->inlines[total_inlines - 1].return_deopt_idx = return_deopt_idx(tc, invoke_ins);
+    inliner->inlines[total_inlines - 1].unreachable = 0;
+    inliner->inlines[total_inlines - 1].deopt_named_used_bit_field =
+        inlinee->deopt_named_used_bit_field;
     inliner->num_inlines = total_inlines;
 
     /* Create/update per-specialization local and lexical type maps. */
@@ -439,6 +534,18 @@ static void merge_graph(MVMThreadContext *tc, MVMSpeshGraph *inliner,
             inlinee->lexical_types ? inlinee->lexical_types : inlinee->sf->body.lexical_types,
             inlinee->num_lexicals * sizeof(MVMuint16));
 
+    /* Merge unreachable handlers array if needed. */
+    if (inliner->unreachable_handlers || inlinee->unreachable_handlers) {
+        MVMuint32 total_handlers = inliner->num_handlers + inlinee->num_handlers;
+        MVMint8 *new_uh = MVM_spesh_alloc(tc, inliner, total_handlers);
+        if (inliner->unreachable_handlers)
+            memcpy(new_uh, inliner->unreachable_handlers, inliner->num_handlers);
+        if (inlinee->unreachable_handlers)
+            memcpy(new_uh + inliner->num_handlers, inlinee->unreachable_handlers,
+                inlinee->num_handlers);
+        inliner->unreachable_handlers = new_uh;
+    }
+
     /* Merge handlers from inlinee. */
     if (inlinee->num_handlers) {
         MVMuint32 total_handlers = inliner->num_handlers + inlinee->num_handlers;
@@ -448,11 +555,20 @@ static void merge_graph(MVMThreadContext *tc, MVMSpeshGraph *inliner,
         for (i = inliner->num_handlers; i < total_handlers; i++) {
             inliner->handlers[i].block_reg += inliner->num_locals;
             inliner->handlers[i].label_reg += inliner->num_locals;
-            if (inliner->sf != inlinee->sf->body.outer)
-                inliner->handlers[i].inlined_and_not_lexical = 1;
+            if (inliner->handlers[i].inlinee == -1)
+                inliner->handlers[i].inlinee = total_inlines - 1;
+            else
+                inliner->handlers[i].inlinee += orig_inlines;
         }
     }
 
+    /* Insert inline boundary entry into the handlers table. */
+    *inline_boundary_handler = inliner->num_handlers + inlinee->num_handlers;
+    resize_handlers_table(tc, inliner, *inline_boundary_handler + 1);
+    inliner->handlers[*inline_boundary_handler].category_mask = MVM_EX_INLINE_BOUNDARY;
+    inliner->handlers[*inline_boundary_handler].action = 0;
+    inliner->handlers[*inline_boundary_handler].inlinee = total_inlines - 1;
+
     /* If the inliner has handlers in effect at the point of the call that we
      * are inlining, then we duplicate those and place them surrounding the
      * inlinee, but with the goto still pointing to the original location.
@@ -482,10 +598,13 @@ static void merge_graph(MVMThreadContext *tc, MVMSpeshGraph *inliner,
                         /* Only update these to the point we found the invoke
                          * being inlined, so it serves as a snapshot of what
                          * is active. */
-                        if (ann->type == MVM_SPESH_ANN_FH_START)
-                            active[ann->data.frame_handler_index] = 1;
+                        MVMint32 fhidx = ann->data.frame_handler_index;
+                        if (ann->type == MVM_SPESH_ANN_FH_START &&
+                                (!inliner->unreachable_handlers ||
+                                 !inliner->unreachable_handlers[fhidx]))
+                            active[fhidx] = 1;
                         else if (ann->type == MVM_SPESH_ANN_FH_END)
-                            active[ann->data.frame_handler_index] = 0;
+                            active[fhidx] = 0;
                     }
                     ann = ann->next;
                 }
@@ -508,9 +627,9 @@ static void merge_graph(MVMThreadContext *tc, MVMSpeshGraph *inliner,
         /* If we found handlers active at the point of invoke, duplicate them
          * in the handlers table and add annotations. */
         if (active_handlers_at_invoke) {
-            MVMuint32 insert_pos = inliner->num_handlers + inlinee->num_handlers;
+            MVMuint32 insert_pos = inliner->num_handlers + inlinee->num_handlers + 1;
             resize_handlers_table(tc, inliner, insert_pos + active_handlers_at_invoke);
-            for (i = orig_handlers - 1; i >= 0; i--) {
+            for (i = 0; i < orig_handlers; i++) {
                 if (active[i]) {
                     /* Add handler start annotation to first inlinee instruction. */
                     MVMSpeshAnn *new_ann = MVM_spesh_alloc(tc, inliner, sizeof(MVMSpeshAnn));
@@ -547,19 +666,23 @@ static void merge_graph(MVMThreadContext *tc, MVMSpeshGraph *inliner,
     inliner->num_bbs      += inlinee->num_bbs - 1;
     inliner->num_locals   += inlinee->num_locals;
     inliner->num_lexicals += inlinee->num_lexicals;
-    inliner->num_handlers += inlinee->num_handlers + active_handlers_at_invoke;
+    inliner->num_handlers += inlinee->num_handlers + 1 + active_handlers_at_invoke;
 }
 
 /* Tweak the successor of a BB, also updating the target BBs pred. */
 static void tweak_succ(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshBB *bb, MVMSpeshBB *new_succ) {
     if (bb->num_succ == 0) {
+        /* It had no successors, so we'll add one. */
         bb->succ = MVM_spesh_alloc(tc, g, sizeof(MVMSpeshBB *));
         bb->num_succ = 1;
-    }
-    if (bb->num_succ == 1)
         bb->succ[0] = new_succ;
-    else
-        MVM_oops(tc, "Spesh inline: unexpected num_succ");
+    }
+    else {
+        /* Otherwise, we can assume that the first successor is the one to
+         * update; others will be there as a result of control handlers, but
+         * these are always added last. */
+        bb->succ[0] = new_succ;
+    }
     if (new_succ->num_pred == 0) {
         new_succ->pred = MVM_spesh_alloc(tc, g, sizeof(MVMSpeshBB *));
         new_succ->num_pred = 1;
@@ -761,12 +884,13 @@ static void rewrite_args(MVMThreadContext *tc, MVMSpeshGraph *inliner,
                 case MVM_OP_arg_n:
                 case MVM_OP_arg_s:
                 case MVM_OP_arg_o:
-                    /* Arg passer just becomes a set instruction; delete the
-                     * parameter-taking instruction. */
-                    arg_ins->info        = MVM_op_get_op(MVM_OP_set);
-                    arg_ins->operands[0] = ins->operands[0];
-                    MVM_spesh_manipulate_delete_ins(tc, inliner, bb, ins);
-                    MVM_spesh_get_facts(tc, inliner, arg_ins->operands[0])->usages++;
+                    /* Receiver just becomes a set instruction; delete the
+                     * argument passing instruction. */
+                    ins->info = MVM_op_get_op(MVM_OP_set);
+                    ins->operands[1] = arg_ins->operands[1];
+                    MVM_spesh_get_facts(tc, inliner, ins->operands[1])->usages++;
+                    MVM_spesh_manipulate_delete_ins(tc, inliner,
+                        call_info->prepargs_bb, arg_ins);
                     break;
                 case MVM_OP_argconst_i:
                     arg_ins->info        = MVM_op_get_op(MVM_OP_const_i64);
@@ -812,8 +936,6 @@ static void rewrite_args(MVMThreadContext *tc, MVMSpeshGraph *inliner,
             case MVM_OP_arg_n:
             case MVM_OP_arg_s:
             case MVM_OP_arg_o:
-                MVM_spesh_get_facts(tc, inliner, arg_ins->operands[1])->usages--;
-                /* fallthrough */
             case MVM_OP_argconst_i:
             case MVM_OP_argconst_n:
             case MVM_OP_argconst_s:
@@ -840,8 +962,9 @@ static void rewrite_args(MVMThreadContext *tc, MVMSpeshGraph *inliner,
 /* Annotates first and last instruction in post-processed inlinee with start
  * and end inline annotations. */
 static void annotate_inline_start_end(MVMThreadContext *tc, MVMSpeshGraph *inliner,
-                               MVMSpeshGraph *inlinee, MVMint32 idx) {
-    /* Annotate first instruction. */
+                               MVMSpeshGraph *inlinee, MVMint32 idx,
+                               MVMuint32 inline_boundary_handler) {
+    /* Annotate first instruction as an inline start. */
     MVMSpeshAnn *start_ann     = MVM_spesh_alloc(tc, inliner, sizeof(MVMSpeshAnn));
     MVMSpeshBB *bb             = inlinee->entry->succ[0];
     start_ann->next            = bb->first_ins->annotations;
@@ -849,14 +972,36 @@ static void annotate_inline_start_end(MVMThreadContext *tc, MVMSpeshGraph *inlin
     start_ann->data.inline_idx = idx;
     bb->first_ins->annotations = start_ann;
 
+    /* Insert annotation for handler boundary indicator fixup. */
+    start_ann = MVM_spesh_alloc(tc, inliner, sizeof(MVMSpeshAnn));
+    start_ann->next = bb->first_ins->annotations;
+    start_ann->type = MVM_SPESH_ANN_FH_START;
+    start_ann->data.frame_handler_index = inline_boundary_handler;
+    bb->first_ins->annotations = start_ann;
+
     /* Now look for last instruction and annotate it. */
     while (bb) {
         if (!bb->linear_next) {
+            /* Annotate it as the inline end. */
             MVMSpeshAnn *end_ann      = MVM_spesh_alloc(tc, inliner, sizeof(MVMSpeshAnn));
             end_ann->next             = bb->last_ins->annotations;
             end_ann->type             = MVM_SPESH_ANN_INLINE_END;
             end_ann->data.inline_idx  = idx;
             bb->last_ins->annotations = end_ann;
+
+            /* Insert annotation for handler boundary fixup; we add the end
+             * one that is needed and also a dummy goto one to keep things
+             * that want all three happy. */
+            end_ann = MVM_spesh_alloc(tc, inliner, sizeof(MVMSpeshAnn));
+            end_ann->next = bb->last_ins->annotations;
+            end_ann->type = MVM_SPESH_ANN_FH_END;
+            end_ann->data.frame_handler_index = inline_boundary_handler;
+            bb->last_ins->annotations = end_ann;
+            end_ann = MVM_spesh_alloc(tc, inliner, sizeof(MVMSpeshAnn));
+            end_ann->next = bb->last_ins->annotations;
+            end_ann->type = MVM_SPESH_ANN_FH_GOTO;
+            end_ann->data.frame_handler_index = inline_boundary_handler;
+            bb->last_ins->annotations = end_ann;
         }
         bb = bb->linear_next;
     }
@@ -866,9 +1011,11 @@ static void annotate_inline_start_end(MVMThreadContext *tc, MVMSpeshGraph *inlin
 void MVM_spesh_inline(MVMThreadContext *tc, MVMSpeshGraph *inliner,
                       MVMSpeshCallInfo *call_info, MVMSpeshBB *invoke_bb,
                       MVMSpeshIns *invoke_ins, MVMSpeshGraph *inlinee,
-                      MVMCode *inlinee_code) {
+                      MVMStaticFrame *inlinee_sf, MVMSpeshOperand code_ref_reg) {
     /* Merge inlinee's graph into the inliner. */
-    merge_graph(tc, inliner, inlinee, inlinee_code, invoke_ins);
+    MVMuint32 inline_boundary_handler;
+    merge_graph(tc, inliner, inlinee, inlinee_sf, invoke_ins, code_ref_reg,
+        &inline_boundary_handler);
 
     /* If we're profiling, note it's an inline. */
     if (inlinee->entry->linear_next->first_ins->info->opcode == MVM_OP_prof_enterspesh) {
@@ -886,8 +1033,11 @@ void MVM_spesh_inline(MVMThreadContext *tc, MVMSpeshGraph *inliner,
      * appropriate slots. */
     rewrite_args(tc, inliner, inlinee, invoke_bb, call_info);
 
-    /* Annotate first and last instruction with inline table annotations. */
-    annotate_inline_start_end(tc, inliner, inlinee, inliner->num_inlines - 1);
+    /* Annotate first and last instruction with inline table annotations; also
+     * add annotations for fixing up the handlers table inline boundary
+     * indicators. */
+    annotate_inline_start_end(tc, inliner, inlinee, inliner->num_inlines - 1,
+        inline_boundary_handler);
 
     /* Finally, turn the invoke instruction into a goto. */
     invoke_ins->info = MVM_op_get_op(MVM_OP_goto);
